#include "common_def.h"


/* ### */
/* abs_interval_to_interval : evaluate the smallest interval including
                              the absolute value of an interval */
void abs_interval_to_interval(x, abs_x)
     INTERVAL *x;
     INTERVAL *abs_x;
{
  if(chk_real_incl_interval(0.,x) == TRUE) {
    abs_x->dw = 0.;
    abs_x->up = x->up;
    if( abs_x->up < -(x->dw) )
      abs_x->up = -(x->dw);
  }
  else if(x->up < 0.)
    neg_interval(x, abs_x);
  else
    copy_interval(x, abs_x);
}

/* ### */
/* add_interval_to : add the first interval to the second one (that will
                     contain the final result) */
void add_interval_to(add1, add2)
     INTERVAL *add1, *add2;
{
  INTERVAL tmp;

  sum_interval(add1, add2, &tmp);
  copy_interval(&tmp, add2);
}

/* ### */
/* calc_util_intv_arit:
           compute the values of one_pm_err.up, one_pm_err.dw and
	   number_limits.dw, number_limits.up, necessary to carry out
	   the interval arithmetics according to the 64 bit IEEE
	   standard for floating point arithmetic. For more
	   information, see H. Koch, A. Schenkel and P. Wittwer:
	   "Computer-assisted proofs in analysis and programming in
	   logic: a case study", SIAM Review, Vol. 38, N. 4,
	   pp. 565-604, December 1996 */
void calc_util_intv_arit()
{
  int j;
  double dx=1.;

  for(j=0;j<52;j++)  dx *= .5;
  one_pm_err.dw = 1. - dx;
  one_pm_err.up = 1. + dx;

  number_limits.dw = number_limits.up = 1.;
  for(j=0;j<500;j++) {
    number_limits.dw *= 0.5;
    number_limits.up *= 2.;
  }
}

/* ### */
/* chk_interval : check that the extremes of the interval are in the safe
                  range to do interval arithmetics */
void chk_interval(x)
     INTERVAL *x;
{
  double tmp1, tmp2, tmp3;

  if(x->dw > x->up) {
    printf("An interval has the upper bound smaller than the lower one!!\n");
    exit(1);
  }

  if((tmp1 = x->up) < 0.) tmp1=-tmp1;
  if((tmp2 = x->dw) < 0.) tmp2=-tmp2;
  if(tmp1<tmp2) {
    tmp3=tmp1;  tmp1=tmp2;  tmp2=tmp3;
  }
  if(tmp1 > number_limits.up) {
    printf("The absolute value of an extreme of an interval is too large!!\n");
    exit(2);
  }

  if( (tmp1 < number_limits.dw && tmp1 != 0.) ||
      (tmp2 < number_limits.dw && tmp2 != 0.) ) {
    /* One of the extreme is not in the safe range to do interval
       arithmetics, but in the present case we can enlarge the
       interval so to arrange this problem */
    if(x->dw < 0. && x->dw > -number_limits.dw)
      x->dw = -number_limits.dw;
    if(x->up < 0. && x->up > -number_limits.dw)
      x->up = 0.;
    if(x->dw > 0. && x->dw < number_limits.dw)
      x->dw = 0.;
    if(x->up > 0. && x->up < number_limits.dw)
      x->up = number_limits.dw;
  }
}

/* ### */
/* chk_real_incl_interval : check that a real value (=double) is inclosed
                            in an interval */
char chk_real_incl_interval(r, r_intv)
     double r;
     INTERVAL *r_intv;
{
  char flag;
  if( ( r_intv->dw <= r ) &&  ( r <= r_intv->up ) )
    flag = TRUE;
  else
    flag = FALSE;
  return(flag);
}

/* ### */
/* copy_interval : copy the limit values of the interval x
                   in the interval y */
void copy_interval(x,y)
     INTERVAL *x, *y;
{
  y->dw = x->dw;
  y->up = x->up;
}

/* ### */
/* cv_int_to_interval : convert an integer to an interval */
void cv_int_to_interval(i,r_intv)
     int i;
     INTERVAL *r_intv;
{
  r_intv->dw = r_intv->up = i;
}

/* ### */
/* cv_real_to_interval : convert a real number (=double) to an interval */
void cv_real_to_interval(r, r_intv)
     double r;
     INTERVAL *r_intv;
{
  if(r < 0.) {
    r_intv->dw = r * one_pm_err.up;
    r_intv->up = r * one_pm_err.dw;
  }
  else if(r == 0.)
    r_intv->dw = r_intv->up = 0.;
  else {
    r_intv->dw = r * one_pm_err.dw;
    r_intv->up = r * one_pm_err.up;
  }
  chk_interval(r_intv);
}

/* ### */
/* div_interval :  calculate the division res = nmr / div ,
                   where res, nmr and div are intervals */
void div_interval(nmr,div,res)
     INTERVAL *nmr;
     INTERVAL *div;
     INTERVAL *res;
{
  INTERVAL invdiv;
  inv_interval(div, &invdiv);
  mul_interval(nmr, &invdiv, res);
}

/* ### */
/* inv_interval :  calculate the division res = 1 / div ,
                   where res and div are intervals */
void inv_interval(div, res)
     INTERVAL *div;
     INTERVAL *res;
{
  if(div->dw > 0.) {
    res->dw = 1. / div->up;  res->dw *= one_pm_err.dw;
    res->up = 1. / div->dw;  res->up *= one_pm_err.up;
  }
  else if(div->up < 0.) {
    res->dw = 1. / div->up;  res->dw *= one_pm_err.up;
    res->up = 1. / div->dw;  res->up *= one_pm_err.dw;
  }
  else {
    printf("Error: division by zero!!\n");
    exit(3);
  }
  chk_interval(res);
}

/* mul_interval : product between two intervals */
void mul_interval(fact1, fact2, res)
     INTERVAL *fact1, *fact2, *res;
{
  if(fact1->dw > 0.) {
    if(fact2->dw > 0.) {
      res->dw = fact1->dw * fact2->dw;  res->dw *= one_pm_err.dw;
      res->up = fact1->up * fact2->up;  res->up *= one_pm_err.up;
    }
    else if(fact2->up < 0.) {
      res->dw = fact1->up * fact2->dw;  res->dw *= one_pm_err.up;
      res->up = fact1->dw * fact2->up;  res->up *= one_pm_err.dw;
    }
    else {
      res->dw = fact1->up * fact2->dw;  res->dw *= one_pm_err.up;
      res->up = fact1->up * fact2->up;  res->up *= one_pm_err.up;
    }
  }
  else if(fact1->up < 0.) {
    if(fact2->dw > 0.) {
      res->dw = fact1->dw * fact2->up;  res->dw *= one_pm_err.up;
      res->up = fact1->up * fact2->dw;  res->up *= one_pm_err.dw;
    }
    else if(fact2->up < 0.) {
      res->dw = fact1->up * fact2->up;  res->dw *= one_pm_err.dw;
      res->up = fact1->dw * fact2->dw;  res->up *= one_pm_err.up;
    }
    else {
      res->dw = fact1->dw * fact2->up;  res->dw *= one_pm_err.up;
      res->up = fact1->dw * fact2->dw;  res->up *= one_pm_err.up;
    }
  }
  else {
    if(fact2->dw > 0.) {
      res->dw = fact1->dw * fact2->up;  res->dw *= one_pm_err.up;
      res->up = fact1->up * fact2->up;  res->up *= one_pm_err.up;
    }
    else if(fact2->up < 0.) {
      res->dw = fact1->up * fact2->dw;  res->dw *= one_pm_err.up;
      res->up = fact1->dw * fact2->dw;  res->up *= one_pm_err.up;
    }
    else {
      double tmp1,tmp2;
      tmp1 = fact1->dw * fact2->up;  tmp2 = fact1->up * fact2->dw;
      res->dw = MIN(tmp1,tmp2);  res->dw *= one_pm_err.up;
      tmp1 = fact1->dw * fact2->dw;  tmp2 = fact1->up * fact2->up;
      res->up = MAX(tmp1,tmp2);  res->up *= one_pm_err.up;
    }
  }
  chk_interval(res);
}

/* ### */
/* neg_interval : store in the interval y the limit values of -x,
                  where x is an interval */
void neg_interval(x, y)
     INTERVAL *x;
     INTERVAL *y;
{
  y->dw = - (x->up);
  y->up = - (x->dw);
}

/* ### */
/* sum_interval : sum between two intervals */
void sum_interval(add1, add2, res)
     INTERVAL *add1, *add2, *res;
{
  res->dw = add1->dw + add2->dw;
  if (res->dw > 0.)  res->dw *= one_pm_err.dw;
  else  res->dw *= one_pm_err.up;

  res->up = add1->up + add2->up;
  if (res->up > 0.)  res->up *= one_pm_err.up;
  else  res->up *= one_pm_err.dw;

  chk_interval(res);
}

/* ### */
/* test_intv_different_zero : check at least one the limits of the interval
                              x is different from zero */
char test_intv_different_zero(x)
     INTERVAL x;
{
  if( x.dw != 0. || x.up != 0.)
    return(1);
  else
    return(0);
}
