/* Hints for the author of the present program:
   1) add many more useful comments in order to better explain
      what every single function actually does;
   2) modify the program in order to make it more homogeneous
      to expl_transf.c (for instance, use the structure type
      INTERVAL when is needed, eventually move the functions
      allowing to do interval arithmetics in a separate file,
      and so on). */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>


/* ### */
/* Useful constant parameters */
#define NDIM 1
#define NDOF (NDIM + 1)
#define K 2
#define KAM_NORMAL_ORDER_EXPL 25
#define KAM_NORMAL_ORDER_ESTIM 1500
#define MAXPOLDEG 2
#define NUMMAXAPPROX 100
#define MAXMODK 100000000
#define LOGZERO -1.e4
#define TRUE 1
#define FALSE 0


/* ### */
/* Global variables */
double onem,onep,smallest,biggest,sqrt_smallest,sqrt_biggest,inf_exp_sum_log;


/* ### */
/* Function prototypes */

/* About the estimates on the expansions */
double bound_sup_norm_strip_angles(double fun_bounds[],
				   double coef_fourier_decay,double E_energy,
				   double a_decay, int flag_deg_0_action);
void calc_estimates_ham_chi1(double ham_bounds[][KAM_NORMAL_ORDER_ESTIM+1],
			     double chi1_bounds[], int r,
			     unsigned int flag_expl_calc);
void calc_estimates_ave_ham_chi1(double ham_bounds[][KAM_NORMAL_ORDER_ESTIM+1],
				 double stima_X, int r,
				 double ham_bounds_ave_lin[]);
void calc_estimates_ham_chi2(double ham_bounds[][KAM_NORMAL_ORDER_ESTIM+1],
			     double chi2_bounds[], int r,
			     unsigned int flag_expl_calc);
void calc_estimates_ave_ham_chi2(double ham_bounds[][KAM_NORMAL_ORDER_ESTIM+1],
				 double chi2_bounds[], int r,
				 double ham_bounds_ave_lin[]);
double calc_gamma(int max_mod_k,double omega[][2],int best_k[][NDOF],
		  double small_div[]);
void calc_param_proof(double ham_bounds[][KAM_NORMAL_ORDER_ESTIM+1],
		      double omega[][2],
		      double E_energy, double a_decay, double greek_zeta,
		      double *rho, double *m_nondeg, double *size_pert);
void calc_small_div(int max_mod_k,double omega[][2],
		    int best_k[][NDOF],double small_div[]);
void initialize_estimates_param(double max_norm_der_X,double chi2_bounds[],
				double *E_energy,double *a_decay,
				double *greek_zeta);
void read_frequencies(FILE *ifp,double omega[][2]);
void read_ham_bounds(FILE *ifp,double ham_bounds[][KAM_NORMAL_ORDER_ESTIM+1]);
void read_gen_fun_bounds(FILE *ifp,double chi_bounds[]);
void update_estimates_param_chi1(double chi1_bounds[],int r,double *E_energy,
				 double *a_decay,double *greek_zeta);
void update_estimates_param_chi2(double chi2_bounds[],int r,double *E_energy,
				 double *a_decay,double *greek_zeta);
void write_cont_fract(int max_mod_k,int best_k[][NDOF],
		      double small_div[],FILE *ofp);

/* About the interval arithmetics */
void chk_number(double x);
double divd(double num,double den);
double divu(double num,double den);
double expu(double x);
double logd(double x);
double logu(double x);
double muld(double fact1,double fact2);
double mulu(double fact1,double fact2);
void small_big();
double sqrtd(double x);
double sqrtu(double x);
double sub_log_d(double a,double b);
double subd(double add,double sub);
double subu(double add,double sub);
double sum_log_u(double a,double b);
double sumd(double add1,double add2);
double sumu(double add1,double add2);
void updown();


/* ### */
int main()
{
  int i, l, r, s, best_k[NUMMAXAPPROX][NDOF];
  double ham_bounds[MAXPOLDEG+1][KAM_NORMAL_ORDER_ESTIM+1];
  double ham_bounds_ave_lin[KAM_NORMAL_ORDER_ESTIM+1];
  double chi1_bounds[2], chi2_bounds[2], max_norm_der_X;
  double gamma, omega[NDOF][2], small_div[NUMMAXAPPROX];
  double E_energy, a_decay, greek_zeta;
  double rho, m_nondeg, size_pert, Lambda, sigma, eps_star, tmp;
  FILE *ifp1,*ifp2;
  FILE *ofp;

  /* Calculate some utilities for doing interval arithmetics */
  small_big();
  updown();


  /* REMARK:
     for nearly all the following instructions, the values of
     the relevant quantities involved in the algorithm are
     REPLACED BY THE LOGARITHM OF THE VALUES THEMSELVES.
     This "trick" has been introduced in order to keep all
     the numbers in the safe range to do interval arithmetics! */


  /* For what concerns all the upper bounds on the norm of the
     Hamiltonian terms that have NOT been (preliminarly) explicitly
     calculated, put them equal to an "approximation of the logarithm
     of zero", i.e. a very large (in absolute value) negative
     number */
  for(l=0;l<MAXPOLDEG+1;l++) {
    for(s=KAM_NORMAL_ORDER_EXPL+1;s<=KAM_NORMAL_ORDER_ESTIM;s++) {
      ham_bounds[l][s] = LOGZERO;
    }
  }

  ofp = fopen("util_proof.out","w");
  fprintf(ofp,"%4d %3d\n",KAM_NORMAL_ORDER_ESTIM,MAXPOLDEG);

  /* Read all the upper bounds on the norm of the Hamiltonian terms
     that HAVE been (preliminarly) explicitly calculated */
  ifp1 = fopen("bounds_expl_expans.bin", "rb");
  read_ham_bounds(ifp1,ham_bounds);


  /* Construction of the Kolmogorov's normal form:
     calculate the needed estimates for the Hamiltonians generated by
     the first KAM_NORMAL_ORDER_EXPL steps of the algorithm (i.e., when
     the expansions of the generating functions are, preliminarly,
     explicitly determined) */
  for(r=1;r<=KAM_NORMAL_ORDER_EXPL;r++) {

    read_gen_fun_bounds(ifp1,chi1_bounds);
    fprintf(ofp,"  r = %3d\n",r);
    if(r>1) {
      update_estimates_param_chi1(chi1_bounds,r,
				  &E_energy,&a_decay,&greek_zeta);
      fprintf(ofp,"E_energy = %24.16le\n",E_energy);
      fprintf(ofp,"greek_zeta = %24.16le\n",greek_zeta);
      fprintf(ofp,"a_decay = %24.16le\n",a_decay);
    }
    else
      max_norm_der_X = chi1_bounds[0];
    calc_estimates_ham_chi1(ham_bounds,chi1_bounds,r,TRUE);
    read_ham_bounds(ifp1,ham_bounds);

    read_gen_fun_bounds(ifp1,chi1_bounds);
    if(r>1) {
      update_estimates_param_chi1(chi1_bounds,r,
				  &E_energy,&a_decay,&greek_zeta);
      fprintf(ofp,"E_energy = %24.16le\n",E_energy);
      fprintf(ofp,"greek_zeta = %24.16le\n",greek_zeta);
      fprintf(ofp,"a_decay = %24.16le\n",a_decay);
    }
    calc_estimates_ham_chi1(ham_bounds,chi1_bounds,r,TRUE);
    read_ham_bounds(ifp1,ham_bounds);

    read_gen_fun_bounds(ifp1,chi2_bounds);
    if(r>1)
      update_estimates_param_chi2(chi2_bounds,r,
				  &E_energy,&a_decay,&greek_zeta);
    else
      initialize_estimates_param(max_norm_der_X,chi2_bounds,
				 &E_energy,&a_decay,&greek_zeta);
    fprintf(ofp,"E_energy = %24.16le\n",E_energy);
    fprintf(ofp,"greek_zeta = %24.16le\n",greek_zeta);
    fprintf(ofp,"a_decay = %24.16le\n",a_decay);
    calc_estimates_ham_chi2(ham_bounds,chi2_bounds,r,TRUE);
    read_ham_bounds(ifp1,ham_bounds);

  }

  /* Read the parameter m_nondeg, that is related to the nondegeneracy
     of the average of the quadratic part of the Hamiltonian (in the
     last approximation that HAS been (preliminarly) explicitly
     calculated) */
  fread(&m_nondeg,sizeof(double),1,ifp1);
  m_nondeg = logd(m_nondeg);
  fclose(ifp1);

  /* Handling of the frequencies related to the invariant tori we are
     interested in */
  ifp2=fopen("freq_intervals.inp","r");
  read_frequencies(ifp2,omega);
  fclose(ifp2);
  calc_small_div(MAXMODK,omega,best_k,small_div);
  gamma = calc_gamma(MAXMODK,omega,best_k,small_div);
  for (i=1; i<NDOF; i++)
    gamma = divd( gamma , NDOF*onem );
  printf("gamma = %24.16le\n",gamma);


  /* Construction of the Kolmogorov's normal form:
     calculate the needed upper bounds for all the terms of the
     Hamiltonians generated by the algorithm steps starting from
     KAM_NORMAL_ORDER_EXPL +1 to KAM_NORMAL_ORDER_ESTIM.
     To this aim, it is enough to iterate a suitable scheme of
     estimates */
  for(r=KAM_NORMAL_ORDER_EXPL+1;r<=KAM_NORMAL_ORDER_ESTIM;r++) {

    /* Calculation of the smallest divisor that can be introduced
       by the KAM normalization at order r */
    i = 0;
    while(small_div[i] > 0.) {
      if( best_k[i][0]+best_k[i][1] > r*K ) {
        i--;
	break;
      }
      i++;
    }
    if( small_div[i] == 0. ) {
      printf("The accuracy in the knowledge of the ");
      printf("small divisors is too low!!\n");
      abort();
    }
    tmp = ham_bounds[0][r];
    if(r == KAM_NORMAL_ORDER_EXPL+1) {
      for(s=1;s<r;s++) {
        tmp = sum_log_u(tmp,ham_bounds[0][s]);
      }
      ham_bounds_ave_lin[r] = ham_bounds[1][0];
    }
    for(l=0;l<NDOF;l++)
      printf("%8d",best_k[i][l]);
    printf("  %24.16le\n",small_div[i]);

    chi1_bounds[0] = sumu( sumu( tmp , logu( divu(onep,small_div[i]) ) ) ,
			  logu(r*K*onep) );
    chi1_bounds[1] = subu( ham_bounds_ave_lin[r] , m_nondeg );
    printf("chi1_bounds[0] = %24.16le\n",chi1_bounds[0]);
    printf("chi1_bounds[1] = %24.16le\n",chi1_bounds[1]);
    update_estimates_param_chi1(chi1_bounds,r,&E_energy,&a_decay,&greek_zeta);
    fprintf(ofp,"  r = %3d\n",r);
    fprintf(ofp,"E_energy = %24.16le\n",E_energy);
    fprintf(ofp,"greek_zeta = %24.16le\n",greek_zeta);
    fprintf(ofp,"a_decay = %24.16le\n",a_decay);
    calc_estimates_ave_ham_chi1(ham_bounds,chi1_bounds[0],r,
				ham_bounds_ave_lin);
    calc_estimates_ham_chi1(ham_bounds,chi1_bounds,r,FALSE);

    if(r == KAM_NORMAL_ORDER_EXPL+1) {
      for(s=1;s<r;s++) {
        ham_bounds[1][r] = sum_log_u(ham_bounds[1][r],ham_bounds[1][s]);
	ham_bounds[1][s] = LOGZERO;
      }
    }
    chi2_bounds[1] = sumu( ham_bounds[1][r] ,
			   logu( divu(onep,small_div[i]) ) );
    chi2_bounds[0] = sumu( chi2_bounds[1] , logu(r*K*onep) );
    printf("chi2_bounds[0] = %24.16le\n",chi2_bounds[0]);
    printf("chi2_bounds[1] = %24.16le\n",chi2_bounds[1]);
    update_estimates_param_chi2(chi2_bounds,r,&E_energy,&a_decay,&greek_zeta);
    fprintf(ofp,"E_energy = %24.16le\n",E_energy);
    fprintf(ofp,"greek_zeta = %24.16le\n",greek_zeta);
    fprintf(ofp,"a_decay = %24.16le\n",a_decay);
    calc_estimates_ave_ham_chi2(ham_bounds,chi2_bounds,r,ham_bounds_ave_lin);
    calc_estimates_ham_chi2(ham_bounds,chi2_bounds,r,FALSE);

    printf("ham_bounds[2][%2d] = %24.16le\n",r,ham_bounds[2][r]);
    m_nondeg = sub_log_d( m_nondeg ,
			  sumu( logu(2.*onep) , ham_bounds[2][r] ) );
    printf("m_nondeg = %24.16le\n", m_nondeg);
    fprintf(ofp,"m_nondeg = %24.16le\n", m_nondeg);

  }


  /* In the following last part of the main function, the 
     applicability of the KAM theorem is checked.
     We refer to the particular statement of the KAM theorem reported
     at page 209 of the Vol. 79 B n.2 (year 1984) of the journal "IL
     NUOVO CIMENTO", in the article written by
     Benettin, Galgani, Giorgilli and Strelcyn:
     "A proof of Kolmogorov's theorem on invariant tori using
     canonical transformations defined by the Lie method". */

  calc_param_proof(ham_bounds,omega,E_energy,a_decay,greek_zeta,
		   &rho,&m_nondeg,&size_pert);

  sigma = mulu( sumu(4*NDOF,onep) , logu(2.*onep) );
  tmp = subu( logu( sumu(NDOF,onep) ) , onem );
  tmp = mulu( sumu(NDOF,onep) , tmp );
  sigma = sumu(sigma,tmp);

  tmp = logu( sumu(4*NDOF,onep) );
  tmp = sumu(tmp,sigma);
  tmp = subu( tmp , logd(gamma) );
  tmp = mulu( tmp , 2*onep );
  Lambda = sumu( logu(2*onep) , tmp );

  eps_star = sumd( muld(6*onep,m_nondeg) , rho );
  eps_star = subd( eps_star , logu(26.*onep) );
  eps_star = subd( eps_star , mulu(2.*onep,Lambda) );
  tmp = subd( rho , logu(13.*onep) );
  tmp = muld( 2*NDOF*onep+6*onep , tmp );
  eps_star = sumu( eps_star , tmp );

  /* REMARK:
     hereafter, we quit the "logarithmic representation" of the
     quantities involved (i.e. the values of the relevant quantities
     are not replaced by their logarithm anymore) */

  fprintf(ofp,"\n\n        ### Proof's summary ###\n");

  fprintf(ofp,"\n  Continued fraction expansion of the frequencies ratio:\n");
  write_cont_fract(MAXMODK,best_k,small_div,ofp);

  fprintf(ofp,"\n  Parameters entering the laws bounding all the terms");
  fprintf(ofp," of the Hamiltonian:\n");
  E_energy = expu(E_energy);
  fprintf(ofp,"E_energy = %24.16le\n",E_energy);
  greek_zeta = expu(greek_zeta);
  fprintf(ofp,"greek_zeta = %24.16le\n",greek_zeta);
  a_decay = expu(a_decay);
  fprintf(ofp,"a_decay = %24.16le\n",a_decay);

  fprintf(ofp,"\n  Values of the parameters appearing in the statement");
  fprintf(ofp," of the theorem:\n");
  fprintf(ofp,"gamma = %24.16le\n",gamma);
  fprintf(ofp,"rho = %24.16le\n",divd(onem,expu(-rho)));
  fprintf(ofp,"m_nondeg = %24.16le\n",divd(onem,expu(-m_nondeg)));

  fprintf(ofp,"\n  Therefore, we claim that:\n");
  if(size_pert < eps_star)
    fprintf(ofp,"the wanted torus (tori) does (do) exist!!\n");
  else {
    fprintf(ofp,"we are not able (yet) to prove that the existence of ");
    fprintf(ofp,"any wanted torus!!\n");
  }

  fprintf(ofp,"\n  Additional information:\n");
  fprintf(ofp,"  (don't worry in case of failure ");
  fprintf(ofp,"of the following calculations)\n");
  eps_star = divd(onem,expu(-eps_star));
  fprintf(ofp,"eps_star = %24.16le\n",eps_star);
  size_pert = expu(size_pert);
  fprintf(ofp,"size_pert = %24.16le\n",size_pert);
  fclose(ofp);

}


/* ### */
double bound_sup_norm_strip_angles(fun_bounds, coef_fourier_increase, E_energy,
				   a_decay, flag_deg_0_action)
     double fun_bounds[], coef_fourier_increase, E_energy, a_decay;
     int flag_deg_0_action;
{
  int s;
  double tmp, effective_decay, bound_sup_norm=LOGZERO;
  printf("Entry bound_sup_norm_strip_angles.\n");
  
  for(s=flag_deg_0_action; s<=KAM_NORMAL_ORDER_ESTIM; s++) {
    tmp = sumu( fun_bounds[s] , mulu(s*onep,coef_fourier_increase) );
    bound_sup_norm = sum_log_u(bound_sup_norm,tmp);
  }

  effective_decay = sumu(a_decay, coef_fourier_increase);
  tmp = divu( onep , sub_log_d(0.,effective_decay) );
  tmp = sumu( tmp ,
	      mulu( (KAM_NORMAL_ORDER_ESTIM+1)*onep , effective_decay ) );
  tmp = sumu( tmp , E_energy );
  bound_sup_norm = sum_log_u(bound_sup_norm,tmp);

  printf("Exit bound_sup_norm_strip_angles.\n");
  return(bound_sup_norm);
}


/* ### */
void calc_estimates_ave_ham_chi1(ham_bounds,stima_X,r,ham_bounds_ave_lin)
     double ham_bounds[MAXPOLDEG+1][KAM_NORMAL_ORDER_ESTIM+1];
     double stima_X,ham_bounds_ave_lin[KAM_NORMAL_ORDER_ESTIM+1];
     int r;
{
  int s,init_s,fin_s;
  double log_fact_coef,add_bound,tmp;
  printf("Entry calc_estimates_ave_ham_chi1.\n");

  if(r == KAM_NORMAL_ORDER_EXPL+1)
    init_s=r+1;
  else
    init_s=2*r-1;

  fin_s=2*r;
  if(fin_s>KAM_NORMAL_ORDER_ESTIM)
    fin_s=KAM_NORMAL_ORDER_ESTIM;

  for(s=init_s;s<=fin_s;s++) {
    ham_bounds_ave_lin[s] = ham_bounds[1][s];
  }

  fin_s=r;
  if(fin_s>KAM_NORMAL_ORDER_ESTIM-r)
    fin_s=KAM_NORMAL_ORDER_ESTIM-r;
  tmp = sumu( logu(2.*onep) , stima_X );

  for(s=1;s<=fin_s;s++) {
    add_bound = sumu( ham_bounds[2][s] , tmp );
    ham_bounds_ave_lin[s+r] = sum_log_u(ham_bounds_ave_lin[s+r],add_bound);
  }

  printf("Exit calc_estimates_ave_ham_chi1.\n");
}


/* ### */
void calc_estimates_ave_ham_chi2(ham_bounds,chi2_bounds,r,ham_bounds_ave_lin)
     double ham_bounds[MAXPOLDEG+1][KAM_NORMAL_ORDER_ESTIM+1];
     double chi2_bounds[2],ham_bounds_ave_lin[KAM_NORMAL_ORDER_ESTIM+1];
     int r;
{
  double add_bound,tmp;
  printf("Entry calc_estimates_ave_ham_chi2.\n");

  if(2*r<KAM_NORMAL_ORDER_ESTIM) {
    tmp = sum_log_u( chi2_bounds[0] ,
		     sumu( logu(2.*r*onep) , chi2_bounds[1] ) );
    add_bound = sumu( ham_bounds[1][r] , tmp );
    ham_bounds_ave_lin[2*r] = sum_log_u(ham_bounds_ave_lin[2*r],add_bound);
  }

  printf("Exit calc_estimates_ave_ham_chi2.\n");
}


/* ### */
void calc_estimates_ham_chi1(ham_bounds,chi1_bounds,r,flag_expl_calc)
     double ham_bounds[MAXPOLDEG+1][KAM_NORMAL_ORDER_ESTIM+1],chi1_bounds[2];
     int r;
     unsigned int flag_expl_calc;
{
  int j,jmax,l,s,fin_s;
  double add_bound,log_fact_coef,chi1_sum_bound;
  printf("Entry calc_estimates_ham_chi1.\n");

  chi1_sum_bound = sum_log_u(chi1_bounds[0],chi1_bounds[1]);
  fin_s=KAM_NORMAL_ORDER_ESTIM-r;

  for(s=fin_s;s>=0;s--) {
    for(l=1;l<=MAXPOLDEG;l++) {
      jmax=(KAM_NORMAL_ORDER_ESTIM-s)/r;
      if(l<jmax)  jmax=l;
      add_bound = ham_bounds[l][s];
      for(j=1;j<=jmax;j++) {
	log_fact_coef = logu( divu((l+1-j)*onep,j*onem) );
	add_bound = sumu( add_bound , sumu(log_fact_coef,chi1_sum_bound) );
	ham_bounds[l-j][s+j*r] = sum_log_u(ham_bounds[l-j][s+j*r],add_bound);
      }
    }
  }

  if(flag_expl_calc==FALSE) {
    if(r==KAM_NORMAL_ORDER_EXPL+1) {
      for(s=1;s<=KAM_NORMAL_ORDER_EXPL;s++)  ham_bounds[0][s]=LOGZERO;
    }
    ham_bounds[0][r]=LOGZERO;
  }

  printf("Exit calc_estimates_ham_chi1.\n");
}


/* ### */
void calc_estimates_ham_chi2(ham_bounds,chi2_bounds,r,flag_expl_calc)
     double ham_bounds[MAXPOLDEG+1][KAM_NORMAL_ORDER_ESTIM+1],chi2_bounds[2];
     int r;
     unsigned int flag_expl_calc;
{
  int j,jmax,l,s,init_s;
  double add_bound,log_fact_coef,tmpcoef0,tmpcoef1;
  printf("Entry calc_estimates_ham_chi2.\n");

  for(l=0;l<=MAXPOLDEG;l++) {

    if(flag_expl_calc==TRUE) {
      if(l==0)  init_s=1;
      else init_s=0;
    }
    else if(r==KAM_NORMAL_ORDER_EXPL+1) {
      if(l==0)  init_s=r+1;
      else init_s=0;
    }
    else {
      if(l==0)  init_s=r+1;
      else if(l==1)  init_s=r;
      else init_s=0;
    }

    if(l==0)
      tmpcoef0 = LOGZERO;
    else
      tmpcoef0 = sumu( logu(l*onep) , chi2_bounds[0] );

    for(s=KAM_NORMAL_ORDER_ESTIM-r;s>=init_s;s--) {

      jmax=(KAM_NORMAL_ORDER_ESTIM-s)/r;
      add_bound = ham_bounds[l][s];

      for(j=1;j<=jmax;j++) {
	if(s==0 && j==1)  tmpcoef1=LOGZERO;
	else
	  tmpcoef1 = sumu( logu( mulu(2.,(s+(j-1)*r)*onep) ) ,
			   chi2_bounds[1] );
	log_fact_coef = logu( divu(1,j*onem) );
	add_bound = sumu( add_bound,
			  sumu( sum_log_u(tmpcoef0,tmpcoef1) ,
				log_fact_coef) );
	ham_bounds[l][s+j*r] = sum_log_u(ham_bounds[l][s+j*r],add_bound);
      }

    }

  }

  if(flag_expl_calc==FALSE) {
    if(r==KAM_NORMAL_ORDER_EXPL+1) {
      for(s=1;s<=KAM_NORMAL_ORDER_EXPL;s++)  ham_bounds[1][s]=LOGZERO;
    }
    ham_bounds[1][r]=LOGZERO;
  }

  printf("Exit calc_estimates_ham_chi2.\n");
}


/* ### */
/* calc_gamma: the present function determines a value of the gamma
               parameter appearing in the Diophantine condition
	       related to a frequency vector (omega0,omega1), i.e.
	       | k0 * omega0 + k1 * omega1 | > gamma / (| k0 | + | k1 |) .
	       The present function exploits the continued fraction
	       expansion, thus it cannot be trivially adapted to the
	       general case of a frequency vector having more than
	       two components. Moreover, the computational algorithm is
	       based on the fact that we consider a special class
	       of frequency vector (omega0,omega1) such that the final
	       part of the continued fraction expansion of the ratio
	       | omega0 / omega1 | is given by an infinite repetition
	       of the number 1. */
double calc_gamma(max_mod_k, omega, best_k, small_div)
     int max_mod_k;
     double omega[2][2];
     int best_k[][2];
     double small_div[];
{
  int i,j,flag;
  double x[2],abs_omega[2][2],tmp,golden_mean[2],gamma;
  printf("Entry calc_gamma.\n");

  for(i=0;i<2;i++) {
    if(omega[i][1]<0.) {
      abs_omega[i][0] = fabs(omega[i][1]);
      abs_omega[i][1] = fabs(omega[i][0]);
    }
    else if(omega[i][0]>0.) {
      abs_omega[i][0] = fabs(omega[i][0]);
      abs_omega[i][1] = fabs(omega[i][1]);
    }
    else {
      printf("0 belongs to an interval of frequencies!!\n");
      abort();
    }
  }

  if( (abs_omega[0][0]+abs_omega[0][1]) >=
      (abs_omega[1][0]+abs_omega[1][1]) ) {
    for(i=0;i<2;i++) {
      tmp = abs_omega[0][i];
      abs_omega[0][i] = abs_omega[1][i];
      abs_omega[1][i] = tmp;
    }
  }

  gamma = small_div[0];
  i=0;
  while(small_div[i] > 0.) {
    for(j=0;j<NDOF;j++)
      printf("%8d",best_k[i][j]);
    printf("  %24.16le\n",small_div[i]);
    tmp = muld( small_div[i] , (best_k[i][0] + best_k[i][1]) * onem );
    if(gamma > tmp) {
      gamma = tmp;
      printf("Update of the value of gamma = %24.16le\n",gamma);
    }
    i++;
  }

  if( (best_k[i-1][0]+best_k[i-1][1]<MAXMODK) && (i<NUMMAXAPPROX-1) ) {
    for(j=0;j<NDOF;j++)
      printf("%8d",best_k[i][j]);
    printf("\n");
    for(j=0;j<NDOF;j++)
      printf("%8d",best_k[i+1][j]);
    printf("\n");
    golden_mean[0] = divd( sumd( sqrtd(5.*onem) , onem ) , 2.*onep);
    golden_mean[1] = divu( sumu( sqrtu(5.*onep) , onep ) , 2.*onem);
    x[0] = sumd( muld(best_k[i+1][1]*onem,golden_mean[0]) ,
		 best_k[i][1]*onem );
    x[0] = divd( x[0] , sumu( mulu( best_k[i+1][0]*onep , golden_mean[1] ) ,
			      best_k[i][0]*onep ) );
    x[1] = sumu( mulu(best_k[i+1][1]*onep,golden_mean[1]) ,
		 best_k[i][1]*onep );
    x[1] = divu( x[1] , sumd( muld( best_k[i+1][0]*onem , golden_mean[0] ) ,
			      best_k[i][0]*onem ) );
    printf("%24.16le %24.16le\n",x[0],x[1]);
    if(i%2 == 0)
      small_div[i] = subd( muld( best_k[i][0] * onem , x[0] ) ,
			   best_k[i][1] * onep );
    else
      small_div[i] = subd( best_k[i][1] * onem ,
			   mulu( best_k[i][0] * onep , x[1] ) );
    small_div[i] = muld(small_div[i],abs_omega[1][0]);
    tmp = muld( small_div[i] , (best_k[i][0] + best_k[i][1]) * onem );
    if(gamma > tmp) {
      gamma = tmp;
      printf("Update of the value of gamma = %24.16le\n",gamma);
    }
    small_div[i] = 0.;
    /* We now make use of the fact that, starting from this point,
       continued fraction expansion is given by an infinite
       repetition of the number 1 */
    tmp = divd(abs_omega[1][0],golden_mean[1]);
    if(gamma > tmp) {
      gamma = tmp;
      printf("Update of the value of gamma = %24.16le\n",gamma);
    }

  }
  else {
    printf("The parameter gamma cannot be rightly defined!!\n");
    abort();
  }

  printf("Exit calc_gamma.\n");
  return(gamma);
}


/* ### */
void calc_param_proof(ham_bounds,omega,E_energy,a_decay,greek_zeta,
		      rho,m_nondeg,size_pert)
     double ham_bounds[MAXPOLDEG+1][KAM_NORMAL_ORDER_ESTIM+1], omega[NDOF][2];
     double E_energy,a_decay,greek_zeta;
     double *rho, *m_nondeg, *size_pert;
{
  int i, l;
  double action_radius, coef_fourier_increase, bound_sup_norm_ham;
  double log_2, log_pot2, tmp;
  printf("Entry calc_param_proof.\n");

  log_pot2 = 0.;
  log_2 = log(2.);

  do {

    log_pot2 += log_2;
    action_radius = -greek_zeta-log_pot2;

    tmp = subd( sub_log_d(-a_decay,0.) , log_pot2 );
    if( tmp < logu(subu(onep,1.)) ) {
      printf("The wanted value of rho has not been found\n");
      abort();
    }
    coef_fourier_increase = sum_log_u( 0. , tmp );

    bound_sup_norm_ham = LOGZERO;
    for(i=0;i<NDOF;i++) {
      if(omega[i][0] > 0.)
	tmp = logu(omega[i][1]);
      else
	tmp = logu(-omega[i][0]);
      bound_sup_norm_ham = sum_log_u( bound_sup_norm_ham , tmp );
    }
    bound_sup_norm_ham = sumu( bound_sup_norm_ham , action_radius );

    for(l=0;l<=MAXPOLDEG;l++) {
      tmp = bound_sup_norm_strip_angles(ham_bounds[l], coef_fourier_increase,
					E_energy, a_decay, l==0);
      if(l == 0)
	*size_pert = tmp;
      else if(l == 1)
	if(*size_pert < tmp)
	  *size_pert = tmp;
      tmp = sumu( tmp , mulu(l*onep,action_radius) );
      bound_sup_norm_ham = sum_log_u( bound_sup_norm_ham , tmp );
    }

  } while(bound_sup_norm_ham >= 0.);

  *rho = action_radius;
  tmp = logd(divd(coef_fourier_increase,K*onem));
  if(*rho > tmp)
    *rho = tmp;
  if(*rho >= 0.)
    *rho = -0.1;

  tmp = bound_sup_norm_strip_angles(ham_bounds[2], coef_fourier_increase,
				    E_energy, a_decay, FALSE);
  tmp = - sumu( tmp , logu(2*onep) );
  if(*m_nondeg > tmp)
    *m_nondeg = tmp;
  if(*m_nondeg >= 0.)
    *m_nondeg = -0.1;

  printf("Exit calc_param_proof.\n");
}


/* ### */
/* calc_small_div: the present function calculates the best
                   approximants of a given frequency vector
                   (omega0,omega1) (i.e. the integer vectors (k0,k1)
                   corresponding to the minimum value of j0 * omega0 +
                   j1 * omega1 for all integers j0, j1 such that
                   |j0|+|j1|<|k0|+|k1|).
		   The present function exploits the continued
		   fraction expansion, thus it cannot be trivially
		   adapted to the general case of a frequency vector
		   having more than two components. */
void calc_small_div(max_mod_k, omega, best_k, small_div)
     int max_mod_k;
     double omega[2][2];
     int best_k[][2];
     double small_div[];
{
  int i,j,flag_change,flag_OK,a[NUMMAXAPPROX];
  double x[2],abs_omega[2][2],tmp,irrat_quad_num[2],repr_appr_freq[2];
  printf("Entry calc_small_div.\n");

  for(i=0;i<2;i++) {
    if(omega[i][1]<0.) {
      abs_omega[i][0] = fabs(omega[i][1]);
      abs_omega[i][1] = fabs(omega[i][0]);
    }
    else if(omega[i][0]>0.) {
      abs_omega[i][0] = fabs(omega[i][0]);
      abs_omega[i][1] = fabs(omega[i][1]);
    }
    else {
      printf("0 belongs to an interval of frequencies!!\n");
      abort();
    }
  }

  if( (abs_omega[0][0]+abs_omega[0][1]) <
      (abs_omega[1][0]+abs_omega[1][1]) ) {
    flag_change = FALSE;
  }
  else {
    flag_change = TRUE;
    for(i=0;i<2;i++) {
      tmp = abs_omega[0][i];
      abs_omega[0][i] = abs_omega[1][i];
      abs_omega[1][i] = tmp;
    }
  }

  i=0;
  x[0] = divd(abs_omega[0][0],abs_omega[1][1]);
  x[1] = divu(abs_omega[0][1],abs_omega[1][0]);
  best_k[0][0] = 1;  best_k[0][1] = 0;

  while( (best_k[i][0]+best_k[i][1]<max_mod_k) && (i<NUMMAXAPPROX-2) ) {

    if(i%2 == 0)
      small_div[i] = subd( muld( best_k[i][0] * onem , abs_omega[0][0] ) ,
			   mulu( best_k[i][1] * onep , abs_omega[1][1] ) );
    else
      small_div[i] = subd( muld( best_k[i][1] * onem , abs_omega[1][0] ) ,
			   mulu( best_k[i][0] * onep , abs_omega[0][1] ) );

    i++;
    tmp = x[0];
    x[0] = divd(onem,x[1]);
    x[1] = divu(onep,tmp);
    a[i] = floor(x[1]);
    printf("%lf %lf\n",x[0],x[1]);
    printf("%3d %3d\n",((int) floor(x[0])),a[i]);

    if(((int) floor(x[0])) != a[i]) {
      flag_OK = FALSE;
      irrat_quad_num[0] = divd( subd( sqrtd(5.*onem) , onep ) , 2.*onep);
      irrat_quad_num[1] = divu( subu( sqrtu(5.*onep) , onem ) , 2.*onem);
      tmp = divu( onep , subd(x[1],((int) floor(x[1]))*onep) );
      irrat_quad_num[0] = sumd( muld( ((int) floor(2.*tmp)) , onem ) ,
				irrat_quad_num[0] );
      irrat_quad_num[1] = sumu( mulu( ((int) floor(2.*tmp)) , onep ) ,
				irrat_quad_num[1] );
      best_k[i][0] = a[i]*best_k[i-1][0]+best_k[i-2][0];
      best_k[i][1] = a[i]*best_k[i-1][1]+best_k[i-2][1];
      repr_appr_freq[0] = sumd( muld( best_k[i][1]*onem , irrat_quad_num[0] ) ,
			       best_k[i-1][1]*onem
			       );
      repr_appr_freq[0] = divd( repr_appr_freq[0] ,
			       sumu( mulu( best_k[i][0]*onep ,
					   irrat_quad_num[1] ) ,
				     best_k[i-1][0]*onep
				     )
			       );
      repr_appr_freq[1] = sumu( mulu( best_k[i][1]*onep , irrat_quad_num[1] ) ,
			       best_k[i-1][1]*onep
			       );
      repr_appr_freq[1] = divu( repr_appr_freq[1] ,
			       sumd( muld( best_k[i][0]*onem ,
					   irrat_quad_num[0] ) ,
				     best_k[i-1][0]*onem )
			       );
      printf("%24.16le %24.16le\n",repr_appr_freq[0],repr_appr_freq[1]);

      if( ( repr_appr_freq[0] > divu(abs_omega[0][0],abs_omega[1][1]) ) &&
	  ( repr_appr_freq[1] < divd(abs_omega[0][1],abs_omega[1][0]) ) ) {
	a[i+1] = ((int) floor(2.*tmp));
	best_k[i+1][0] = a[i+1]*best_k[i][0]+best_k[i-1][0];
	best_k[i+1][1] = a[i+1]*best_k[i][1]+best_k[i-1][1];
	break;
      }
      else {
	printf("A Diophantine approximation of the frequency ratio in the\n");
	printf("interval of values initially defined has not been found!!\n");
	abort();
      }

    }

    if(i == 1) {
      best_k[1][0] = a[1];  best_k[1][1] = 1;
    }
    else {
      best_k[i][0] = a[i]*best_k[i-1][0]+best_k[i-2][0];
      best_k[i][1] = a[i]*best_k[i-1][1]+best_k[i-2][1];
    }
    x[0] = subd(x[0],a[i]*onem);
    x[1] = subu(x[1],a[i]*onep);

  }
  small_div[i] = 0.;

  i=0;
  while(small_div[i]>0.) {
    int tmpint;
    tmpint = best_k[i][1-flag_change];
    best_k[i][0] = best_k[i][flag_change];
    best_k[i][1] = tmpint;
    i++;
  }

  printf("Exit calc_small_div.\n");
}


/* ### */
void initialize_estimates_param(max_norm_der_X,chi2_bounds,
				E_energy,a_decay,greek_zeta)
     double max_norm_der_X,chi2_bounds[2],*E_energy,*a_decay,*greek_zeta;
{
  double tmp;
  printf("Entry initialize_estimates_param.\n");

  *a_decay = sumu( logu(K*onep) , chi2_bounds[1] );
  if(*a_decay < chi2_bounds[0])  *a_decay = chi2_bounds[0];

  *E_energy = mulu( 2. , subu(max_norm_der_X,*a_decay) );
  *greek_zeta = subu(*a_decay,max_norm_der_X);


  tmp = divu( KAM_NORMAL_ORDER_ESTIM * onep ,
	      logd(KAM_NORMAL_ORDER_ESTIM * onem) );
  tmp = logu(tmp);
  tmp = subu(tmp,onem);
  *E_energy = sumu(*E_energy,tmp);

  tmp = logu(KAM_NORMAL_ORDER_ESTIM * onep);
  tmp = divu(tmp, KAM_NORMAL_ORDER_ESTIM * onem);
  *a_decay = sumu(tmp,*a_decay);

  printf("E_energy = %24.16le\n",*E_energy);
  printf("greek_zeta = %24.16le\n",*greek_zeta);
  printf("a_decay = %24.16le\n",*a_decay);

  printf("Exit initialize_estimates_param.\n");
}


/* ### */
void read_frequencies(ifp,omega)
     FILE *ifp;
     double omega[NDOF][2];
{
  int i;
  printf("Entry read_frequencies.\n");

  for(i=0;i<NDOF;i++) {
    fscanf(ifp,"%lf %lf",&omega[i][0],&omega[i][1]);
  }

  printf("Exit read_frequencies.\n");
}


/* ### */
void read_gen_fun_bounds(ifp,chi_bounds)
     FILE *ifp;
     double chi_bounds[2];
{
  int i;
  printf("Entry read_gen_fun_bounds.\n");

  for(i=0;i<2;i++) {
    fread(&chi_bounds[i],sizeof(double),1,ifp);
    printf("%24.16le\n",chi_bounds[i]);
    if(chi_bounds[i]==0.)  chi_bounds[i]=LOGZERO;
    else  chi_bounds[i]=logu(chi_bounds[i]);
  }

  printf("Exit read_gen_fun_bounds.\n");
}


/* ### */
void read_ham_bounds(ifp,ham_bounds)
     FILE *ifp;
     double ham_bounds[MAXPOLDEG+1][KAM_NORMAL_ORDER_ESTIM+1];
{
  int l,s;
  printf("Entry read_ham_bounds.\n");

  for(l=0;l<=MAXPOLDEG;l++) {
    for(s=(l==0);s<=KAM_NORMAL_ORDER_EXPL;s++) {
      fread(&ham_bounds[l][s],sizeof(double),1,ifp);
      if(ham_bounds[l][s]==0.)  ham_bounds[l][s]=LOGZERO;
      else  ham_bounds[l][s]=logu(ham_bounds[l][s]);
    }
  }

  printf("Exit read_ham_bounds.\n");
}


/* ### */
void update_estimates_param_chi1(chi1_bounds,r,E_energy,a_decay,greek_zeta)
     double chi1_bounds[2],*E_energy,*a_decay,*greek_zeta;
     int r;
{
  double tmp;
  printf("Entry update_estimates_param_chi1.\n");

  tmp = sum_log_u(chi1_bounds[0],chi1_bounds[1]);
  tmp = sumu( logu(2*onep) , sumu(tmp,*greek_zeta) );
  tmp = subu( tmp , muld(*a_decay,r) );
  tmp = sum_log_u(0.,tmp);

  *greek_zeta = subu(*greek_zeta,tmp);

  tmp = mulu(2.,tmp);
  *E_energy = sumu(*E_energy,tmp);

  printf("E_energy = %24.16le\n",*E_energy);
  printf("greek_zeta = %24.16le\n",*greek_zeta);
  printf("a_decay = %24.16le\n",*a_decay);

  printf("Exit update_estimates_param_chi1.\n");
}


/* ### */
void update_estimates_param_chi2(chi2_bounds,r,E_energy,a_decay,greek_zeta)
     double chi2_bounds[2],*E_energy,*a_decay,*greek_zeta;
     int r;
{
  double tmp;
  printf("Entry update_estimates_param_chi2.\n");

  tmp = sumu( logu(r*K*onep) , chi2_bounds[1] );
  if(tmp < chi2_bounds[0])  tmp = chi2_bounds[0];
  tmp = subu( tmp , muld(*a_decay,r) );
  tmp = sum_log_u(0.,tmp);

  *greek_zeta = sumu(*greek_zeta,tmp);

  tmp = divu(tmp,r*onem);
  *a_decay = sumu( *a_decay , tmp );

  printf("E_energy = %24.16le\n",*E_energy);
  printf("greek_zeta = %24.16le\n",*greek_zeta);
  printf("a_decay = %24.16le\n",*a_decay);

  printf("Exit update_estimates_param_chi2.\n");
}


/* ### */
void write_cont_fract(max_mod_k, best_k, small_div, ofp)
     int max_mod_k;
     int best_k[][2];
     double small_div[];
     FILE *ofp;
{
  int i,j,a,sub;
  printf("Entry write_cont_fract.\n");

  i=0;
  while(small_div[i] > 0.) {
    if(i==0) {
      fprintf(ofp,"%10d",0);
      sub = 0;
    }
    else {
      a = (best_k[i][0] - sub) / best_k[i-1][0];
      fprintf(ofp,"%10d",a);
      sub = best_k[i-1][0];
    }
    if(i%6 == 5)
      fprintf(ofp,"\n");
    i++;
  }

  if( (best_k[i-1][0]+best_k[i-1][1]<MAXMODK) && (i<NUMMAXAPPROX-1) ) {
    a = (best_k[i][0] - sub) / best_k[i-1][0];
    fprintf(ofp,"%10d",a);
    sub = best_k[i-1][0];
    if(i%6 == 5)
      fprintf(ofp,"\n");
    a = (best_k[i+1][0] - sub) / best_k[i][0];
    fprintf(ofp,"%10d",a);
    fprintf(ofp,"\n");
  }
  else {
    printf("It is not possible to rightly define the value of gamma!!\n");
    abort();
  }

  printf("Exit write_cont_fract.\n");
}


/* ### */
/* Beginning of the functions we need to perform interval arithmetics.
   REMARK: in the present program we actually need to calculate the
   upper bound (or the lower bound) of many quantities introduced by
   the algorithm  */
/* ### */


void chk_number(x)
     double x;
{
  x = fabs(x);
  if( (x>smallest && x<biggest) || x==0. ) {
    return;
  }
  printf("The following number is not in the safe range for doing ");
  printf("interval arithmetics!!!\n");
  printf("%24.16le\n",x);
  abort();
}


/* ### */
double divd(num,den)
     double num;
     double den;
{
  double tmp;
  if(den == 0.) {
    printf("Division by zero!!\n");
    abort();
  }
  tmp = num / den;
  if(tmp >= 0.)  tmp *= onem;
  else  tmp *= onep;
  chk_number(tmp);
  return(tmp);
}


/* ### */
double divu(num,den)
     double num;
     double den;
{
  double tmp;
  if(den == 0.) {
    printf("Division by zero!!\n");
    abort();
  }
  tmp = num / den;
  if(tmp >= 0.)  tmp *= onep;
  else  tmp *= onem;
  chk_number(tmp);
  return(tmp);
}


/* ### */
double expu(x)
     double x;
{
  int i,n;
  double res,div,tmp_x,tmp1,tmp2;
  tmp1 = fabs(x);
  if( tmp1 > log(biggest) ) {
    printf("The exp function has been requested to calculate\n");
    printf("a too large (small) number!!\n");
    abort();
  }
  res = 1.;
  n=0;
  while( tmp1 > 0.03 ) {
    tmp1 = divu(tmp1,2.);
    n++;
  }
  div = 1.;
  tmp_x = tmp1;
  if(x >= 0.) {
    for(i=1; (1.+tmp1/div) != 1.; i++) {
      tmp2 = divu(tmp1,div);
      res = sumu(res,tmp2);
      tmp1 = mulu(tmp1,tmp_x);
      tmp2 = sumd(i*onem,1.);
      div = muld(div,tmp2);
    }
    tmp2 = divu(tmp1,div);
    div = subd(1.,tmp_x);
    tmp1 = divu(tmp2,div);
    res = sumu(res,tmp1);
  }
  else {
    for(i=1; (1.+tmp1/div) != 1.; i++) {
      tmp2 = divd(tmp1,div);
      res = sumd(res,tmp2);
      tmp1 = muld(tmp1,tmp_x);
      tmp2 = sumu(i*onep,1.);
      div = mulu(div,tmp2);
    }
    tmp2 = divd(tmp1,div);
    div = subu(1.,tmp_x);
    tmp1 = divd(tmp2,div);
    res = sumd(res,tmp1);
    res = divu(1.,res);
  }
  for(i=0;i<n;i++) {  
    res = mulu(res,res);
  }
  return(res);
}


/* ### */
double logd(x)
     double x;
{
  int i,n;
  double div[2],tmp_x[2],tmp1[2],tmp2[2],res[2],pot2[2];
  if( x <= 0. ) {
    printf("Attempt to calculate the logarithm of a non-positive number!!\n");
    abort();
  }
  if( x < 1.)  tmp_x[0] = divu(1.,x);
  else  tmp_x[0] = x;
  tmp_x[1] = tmp_x[0];
  n=0;
  while( tmp_x[1] > 1.01 ) {
    tmp_x[0] = sqrtd(tmp_x[0]);  tmp_x[1] = sqrtu(tmp_x[1]);
    n++;
  }
  res[0] = res[1] = 0.;
  tmp1[0] = subd(tmp_x[0],1.);   tmp1[1] = subu(tmp_x[1],1.);
  tmp_x[0] = tmp1[0];   tmp_x[1] = tmp1[1];
  div[0] = div[1] = 1.;
  for(i=1; (res[0]+tmp1[1]/div[0]) != res[0];i++) {
    tmp2[0] = divd(tmp1[0],div[1]);  tmp2[1] = divu(tmp1[1],div[0]);
    if(i%2 == 1) {
      res[0] = sumd(res[0],tmp2[0]);  res[1] = sumu(res[1],tmp2[1]);
    }
    else {
      res[0] = subd(res[0],tmp2[1]);  res[1] = subu(res[1],tmp2[0]);
    }
    tmp1[0] = muld(tmp1[0],tmp_x[0]);  tmp1[1] = mulu(tmp1[1],tmp_x[1]);
    div[0] = sumd(i,onem);  div[1] = sumu(i,onep);
  }
  tmp2[0] = divd(tmp1[0],div[1]);  tmp2[1] = divu(tmp1[1],div[0]);
  res[0] = subd(res[0],tmp2[1]);  res[1] = sumu(res[1],tmp2[1]);
  pot2[0] = pot2[1] = 1.;
  for(i=0;i<n;i++) {
    pot2[0] = muld(pot2[0],2.);  pot2[1] = mulu(pot2[1],2.);
  }
  res[0] = muld(res[0],pot2[0]);  res[1] = mulu(res[1],pot2[1]);
  if( x < 1.)  return(-res[1]);
  else  return(res[0]);
}


/* ### */
double logu(x)
     double x;
{
  int i,n;
  double div[2],tmp_x[2],tmp1[2],tmp2[2],res[2],pot2[2];
  if( x <= 0. ) {
    printf("Attempt to calculate the logarithm of a non-positive number!!\n");
    abort();
  }
  if( x < 1.)  tmp_x[0] = divd(1.,x);
  else  tmp_x[0] = x;
  tmp_x[1] = tmp_x[0];
  n=0;
  while( tmp_x[1] > 1.01 ) {
    tmp_x[0] = sqrtd(tmp_x[0]);  tmp_x[1] = sqrtu(tmp_x[1]);
    n++;
  }
  res[0] = res[1] = 0.;
  tmp1[0] = subd(tmp_x[0],1.);   tmp1[1] = subu(tmp_x[1],1.);
  tmp_x[0] = tmp1[0];   tmp_x[1] = tmp1[1];
  div[0] = div[1] = 1.;
  for(i=1; (res[0]+tmp1[1]/div[0]) != res[0];i++) {
    tmp2[0] = divd(tmp1[0],div[1]);  tmp2[1] = divu(tmp1[1],div[0]);
    if(i%2 == 1) {
      res[0] = sumd(res[0],tmp2[0]);  res[1] = sumu(res[1],tmp2[1]);
    }
    else {
      res[0] = subd(res[0],tmp2[1]);  res[1] = subu(res[1],tmp2[0]);
    }
    tmp1[0] = muld(tmp1[0],tmp_x[0]);  tmp1[1] = mulu(tmp1[1],tmp_x[1]);
    div[0] = sumd(i,onem);  div[1] = sumu(i,onep);
  }
  tmp2[0] = divd(tmp1[0],div[1]);  tmp2[1] = divu(tmp1[1],div[0]);
  res[0] = subd(res[0],tmp2[1]);  res[1] = sumu(res[1],tmp2[1]);
  pot2[0] = pot2[1] = 1.;
  for(i=0;i<n;i++) {
    pot2[0] = muld(pot2[0],2.);  pot2[1] = mulu(pot2[1],2.);
  }
  res[0] = muld(res[0],pot2[0]);  res[1] = mulu(res[1],pot2[1]);
  if( x < 1.)  return(-res[0]);
  else  return(res[1]);
}


/* ### */
double muld(fact1,fact2)
     double fact1;
     double fact2;
{
  double tmp;
  tmp = fact1 * fact2;
  if(tmp >= 0.)  tmp *= onem;
  else  tmp *= onep;
  chk_number(tmp);
  return(tmp);
}


/* ### */
double mulu(fact1,fact2)
     double fact1;
     double fact2;
{
  double tmp;
  tmp = fact1 * fact2;
  if(tmp >= 0.)  tmp *= onep;
  else  tmp *= onem;
  chk_number(tmp);
  return(tmp);
}


/* ### */
/* small_big: compute the values of smallest and biggest (i.e. 
              the smallest and biggest positive numbers in the
	      safe range of the representable numbers) necessary
	      to carry out the interval arithmetics according
	      to the 64 bit IEEE standard for floating point
	      arithmetic. For more information, see
	      H. Koch, A. Schenkel and P. Wittwer:
	      "Computer-assisted proofs in analysis and
	      programming in logic: a case study", SIAM Review,
	      Vol. 38, N. 4, pp. 565-604, December 1996 */	
void small_big()
{
  double small=1.;
  double big=1.;
  double sqrt_small=1.;
  double sqrt_big=1.;
  int j;
  for(j=0;j<500;j++) {
    small /= 2.;
    big *= 2.;
  }
  smallest = small;
  biggest = big;
  for(j=0;j<250;j++) {
    sqrt_small /= 2.;
    sqrt_big *= 2.;
  }
  sqrt_smallest = sqrt_small;
  sqrt_biggest = sqrt_big;
  return;
}


/* ### */
double sqrtd(x)
     double x;
{
  if( x < 0. ) {
    printf("Attempt to calculate the square root of a negative number!!\n");
    abort();
  }
  return(sqrt(x) * onem);
}


/* ### */
double sqrtu(x)
     double x;
{
  if( x < 0. ) {
    printf("Attempt to calculate the square root of a negative number!!\n");
    abort();
  }
  return(sqrt(x) * onep);
}


/* ### */
double sub_log_d(a,b)
     double a,b;
{
  double res,tmp;
  if(b>=a) {
    printf("It is not possible to calculate log(exp(a)-exp(b)) ");
    printf("when a <= b !!!\n");
    abort();
  }
  tmp = subu(b,a);
  if(tmp >= inf_exp_sum_log)  res = subd(1.,expu(tmp));
  else  res = onem;
  res = subd( a , logu(divu(onep,res)) );
  return(res);
}


/* ### */
double subd(add,sub)
     double add;
     double sub;
{
  double tmp;
  tmp = add - sub;
  if(tmp >= 0.)  tmp *= onem;
  else  tmp *= onep;
  chk_number(tmp);
  return(tmp);
}


/* ### */
double subu(add,sub)
     double add;
     double sub;
{
  double tmp;
  tmp = add - sub;
  if(tmp >= 0.)  tmp *= onep;
  else  tmp *= onem;
  chk_number(tmp);
  return(tmp);
}


/* ### */
double sum_log_u(a,b)
     double a,b;
{
  double res,tmp;
  if(b>a) {
    tmp=b;  b=a;  a=tmp;
  }
  tmp = subu(b,a);
  if(tmp >= inf_exp_sum_log)  res = sumu(1.,expu(tmp));
  else  res = onep;
  res = sumu(a,logu(res));
  return(res);
}


/* ### */
double sumd(add1,add2)
     double add1;
     double add2;
{
  double tmp;
  tmp = add1 + add2;
  if(tmp >= 0.)  tmp *= onem;
  else  tmp *= onep;
  chk_number(tmp);
  return(tmp);
}


/* ### */
double sumu(add1,add2)
     double add1;
     double add2;
{
  double tmp;
  tmp = add1 + add2;
  if(tmp >= 0.)  tmp *= onep;
  else  tmp *= onem;
  chk_number(tmp);
  return(tmp);
}


/* ### */
/* updown: compute the values of onem and onep necessary
           to carry out the interval arithmetics according
	   to the 64 bit IEEE standard for floating point
	   arithmetic. For more information, see
	   H. Koch, A. Schenkel and P. Wittwer:
	   "Computer-assisted proofs in analysis and
	   programming in logic: a case study", SIAM Review,
	   Vol. 38, N. 4, pp. 565-604, December 1996 */	
void updown()
{
  double dx=1.;
  int j;
  for(j=0;j<52;j++) dx *= .5;
  onem = 1. - dx;
  onep = 1. + dx;
  inf_exp_sum_log = 2. * logu( subu(onep,1.) );
  return;
}


/* ### */
/* End of the functions we need to perform interval arithmetics */
/* ### */
